
rm(list=ls())

library(Rcpp)
library(RcppArmadillo)
library(ggplot2)
library(mvtnorm)
library(tidyr)
library(reshape2)
library(latex2exp)
library(RColorBrewer)
library(pgdraw)
library(flmem)

theme_set(theme_bw(base_size = 14))

cols <- brewer.pal(9, "Set1")


# Data Generation ---------------------------------------------------------
mu_true <- function(t){
  return (- 1 + 0.02 * t^2)
}

xi <- function(t){
  return (rep(1, length(t)))
}

set.seed(1234)
# Choose the number of knots
K <- 20

# Choose the interval for the splines
a <- 1; b <- 20

knots <- seq(a, b, length.out = K)

T_max <- 20
n <- 40
n_tr <- 40

sigma_eta <- 0.75
c2 <- (16 * sqrt(3)/(15 * pi))^2


# Generate some synthetic data
Y <- array(NA, dim = c(n, T_max, n_tr))
p_itk <- array(NA, dim = c(n, T_max, n_tr))
eta_i <- array(NA, dim = n)
for (i in 1:n){
  eta_i[i] <- rnorm(1, 0, sigma_eta)
  for (t in 1:T_max){
    for (k in 1:n_tr){
      p_itk[i,t,k] <- logit_vec(mu_true(t) + eta_i[i] * xi(t))
      Y[i,t,k] <- sample(0:1, 1, FALSE, c(1 - p_itk[i,t,k], p_itk[i,t,k]))
    }
  }
}


# Prepare the data in the right format
# Create a matrix with individuals on the rows, time on the columns, number of
# successes out of the n_tr trials in position {ij}
succ <- apply(Y, c(1, 2), sum)
ni <- rep(dim(Y)[3], dim(Y)[1] * dim(Y)[2]) # number of trials corresponding
data <- melt(succ)
data$ni <- ni
colnames(data) <- c("ind", "time", "succ", "ni")
jit_data <- data
jit_data$time <- jitter(data$time)


ggplot() +
  geom_point(aes(x = time, y = succ/ni),
             col = rgb(0.75, 0.75, 0.75, 0.8), pch = 16, size = 2,
             data = data) +
  geom_line(aes(x = time, y = succ/ni, group = ind), size = 0.5,
            col = rgb(0.75, 0.75, 0.75, 1), data = data) +
  geom_line(aes(x = 1:T_max, y = logit_vec(mu_true(1:T_max)/sqrt(1 + c2 * xi(1:20)^2 * sigma_eta^2)), col = 'True'), size = 1) +
  scale_color_brewer(name="", palette = "Set1") +
  labs(x = 'Number of days', y = 'Probability', title = 'Empirical probabilities of success across trials')


# Binomial Probit NP model with common functional random effects ----------
# logit(\pi_i(t)) = \mu(t) + \xi(t) \eta_i


# Run the MCMC
hypers <- NULL
hypers$P_x <- hypers$P_z <- P_smooth(K + 1)
hypers$a_beta <- hypers$b_beta <- 0.5
hypers$a_gamma <- hypers$b_gamma <- 0.5
hypers$a_eta <- 1
x_pred <- a:b


Niter <- 7500
burnin <- 2500
thin <- 1
samp_size <- (Niter - burnin)/thin


fit_ranef_spl <- bin_logit_randspline(data$succ, B_basis(data$time, knots),
                                      B_basis(data$time, knots), data$ind,
                                      data$ni, data$time,
                                      B_basis(x_pred, knots),
                                      B_basis(x_pred, knots), hypers, Niter,
                                      burnin, thin)

# Compute the quantiles for the estimated mean function
pi_quant_re <- apply(fit_ranef_spl$pi, 2, quantile, prob = c(0.025, 0.975))
mu_quant_re <- apply(fit_ranef_spl$mu, 2, quantile, prob = c(0.025, 0.975))
sig_u_quant_re <- apply(0.8 * sqrt(fit_ranef_spl$sigma2_u), 2, quantile, prob = c(0.025, 0.975))


MISE <- mean(diff(x_pred)[1] * (colMeans(fit_ranef_spl$pi) - logit_vec(mu_true(x_pred)/sqrt(1 + c2 * xi(x_pred)^2 * sigma_eta^2)))^2)
MISE


# Intercept
time <- 1
u_i_quant <- apply(fit_ranef_spl$u[time,,], 1, quantile, probs =  c(0.025, 0.975))
u_summary <- data.frame(ind = factor(1:n),
                        lo = u_i_quant[1,],
                        md = rowMeans(fit_ranef_spl$u[time,,]),
                        hi = u_i_quant[2,])

u_summary <- u_summary[order(u_summary$md),]
u_summary$ind <- factor(u_summary$ind, levels = u_summary$ind)


# Plot the individual random intercepts at t = time
ggplot(u_summary, aes(ymin = lo, ymax = hi, x = ind)) +
  geom_linerange(size = 1) +
  ylab("Random effects distributions") +
  geom_hline(yintercept = 0, linetype = "dashed") +
  geom_point(aes(y = sort(eta_i) * abs(xi(1)), col = 'True'), size = 2) +
  geom_point(aes(y = md, col = 'FLMEM'), size = 2) +
  scale_color_brewer(name="", palette = "Set1") +
  scale_x_discrete(name = 'Individual', breaks = NULL, labels = rep('', n)) +
  theme(plot.title = element_text(hjust = 0.5))


# Plot the population curve on the probability scale
ggplot() +
  geom_point(aes(x = jitter(time), y = succ/ni),
             col = rgb(0.75, 0.75, 0.75, 0.8), pch = 16, size = 2,
             data = data) +
  geom_ribbon(aes(x = x_pred, ymin = pi_quant_re[1,], ymax = pi_quant_re[2,]), fill = cols[1], alpha = 0.4) +
  geom_line(aes(x = x_pred, y = colMeans(fit_ranef_spl$pi), col = 'FLMEM'), size = 1) +
  geom_line(aes(x = x_pred, y = logit_vec(mu_true(x_pred)/sqrt(1 + c2 * xi(x_pred)^2 * sigma_eta^2)), col = 'True'), size = 1, lty = 2) +
  labs(x = 'Number of days', y = 'Probability', title = 'Population Curve') +
  scale_color_brewer(name="", palette = "Set1") +
  theme(legend.title=element_blank())


# Random effects variance estimate
ggplot() +
  geom_line(aes(x = x_pred, y = colMeans(0.8 * sqrt(fit_ranef_spl$sigma2_u))), col = cols[1], size = 1) +
  geom_ribbon(aes(x = x_pred, ymin = sig_u_quant_re[1,], ymax = sig_u_quant_re[2,]), fill = cols[1], alpha = 0.4) +
  geom_line(aes(x = x_pred, y = xi(x_pred) * sigma_eta), col = cols[2], lty = 2, size = 1) +
  labs(x = 'Number of days', y = "Random effects s.d.") +
  scale_color_brewer(name="", palette = "Set1") +
  theme(legend.title=element_blank())

# Plot three individual specific curves
idx1 <- order(colMeans(fit_ranef_spl$eta))[1]
idx2 <- order(colMeans(fit_ranef_spl$eta))[floor(n/2)]
idx3 <- order(colMeans(fit_ranef_spl$eta))[n]

ggplot() +
  geom_point(aes(x = time, y = succ/ni),
             col = rgb(0.75, 0.75, 0.75, 0.8), pch = 16, size = 2,
             data = jit_data) +

  geom_line(aes(x = x_pred, y = colMeans(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx1,])))), col = cols[4], size = 1) +
  geom_point(aes(x = jit_data$time[data$ind == idx1], y = (data$succ/data$ni)[data$ind == idx1]), col = cols[4], pch = 16, size = 2) +
  geom_line(aes(x = jit_data$time[data$ind == idx1], y = (data$succ/data$ni)[data$ind == idx1]), size = 0.5, col = cols[4]) +
  geom_ribbon(aes(x = x_pred, ymin = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx1,])), 2, quantile, prob = 0.025), ymax = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx1,])), 2, quantile, prob = 0.975)), fill = cols[4], alpha = 0.4) +

  geom_line(aes(x = x_pred, y = colMeans(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx2,])))), col = cols[5], size = 1) +
  geom_point(aes(x = jit_data$time[data$ind == idx2], y = (data$succ/data$ni)[data$ind == idx2]), col = cols[5], pch = 16, size = 2) +
  geom_line(aes(x = jit_data$time[data$ind == idx2], y = (data$succ/data$ni)[data$ind == idx2]), size = 0.5, col = cols[5]) +
  geom_ribbon(aes(x = x_pred, ymin = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx2,])), 2, quantile, prob = 0.025), ymax = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx2,])), 2, quantile, prob = 0.975)), fill = cols[5], alpha = 0.4) +

  geom_line(aes(x = x_pred, y = colMeans(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx3,])))), col = "aquamarine3", size = 1) +
  geom_point(aes(x = jit_data$time[data$ind == idx3], y = (data$succ/data$ni)[data$ind == idx3]), col = "aquamarine3", pch = 16, size = 2) +
  geom_line(aes(x = jit_data$time[data$ind == idx3], y = (data$succ/data$ni)[data$ind == idx3]), size = 0.5, col = "aquamarine3") +
  geom_ribbon(aes(x = x_pred, ymin = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx3,])), 2, quantile, prob = 0.025), ymax = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx3,])), 2, quantile, prob = 0.975)), fill = "aquamarine3", alpha = 0.4) +

  labs(x = 'Number of days', y = 'Probability') +
  scale_color_brewer(name="", palette = "Set1") +
  theme(legend.title=element_blank())

