
rm(list=ls())

library(Rcpp)
library(RcppArmadillo)
library(ggplot2)
library(mvtnorm)
library(tidyr)
library(reshape2)
library(latex2exp)
library(RColorBrewer)
library(pgdraw)
library(flmem)

theme_set(theme_bw(base_size = 14))

cols <- brewer.pal(9, "Set1")

data("learnCurve")


# Choose the number of knots
K <- max(learnCurve$time)

# Choose the interval for the splines
a <- min(learnCurve$time); b <- max(learnCurve$time);

knots <- seq(a, b, length.out = K)

T_max <- max(learnCurve$time)
n <- length(unique(learnCurve$ind))
c2 <- (16 * sqrt(3)/(15 * pi))^2 # correction for the population mean


# Plot the observed data
jit_data <- learnCurve
jit_data$time <- jitter(learnCurve$time)
ggplot() +
  geom_point(aes(x = time, y = succ/ni),
             col = rgb(0.75, 0.75, 0.75, 0.8), pch = 16, size = 2,
             data = jit_data) +
  geom_line(aes(x = time, y = succ/ni, group = ind), size = 0.5,
            col = rgb(0.75, 0.75, 0.75, 1), data = jit_data) +
  scale_color_brewer(name="", palette = "Set1") +
  theme(legend.position="none") +
  labs(x = 'Number of days', y = 'Probability', title = 'Empirical probabilities of success across trials')



# Binomial Probit NP model with common functional random effects ----------
# logit(\pi_i(t)) = \mu(t) + \xi(t) \eta_i

# Choose the hyperparameters
hypers <- NULL
hypers$P_x <- hypers$P_z <- P_smooth(K + 1)
hypers$a_beta <- hypers$b_beta <- 0.5
hypers$a_gamma <- hypers$b_gamma <- 0.5
hypers$a_eta <- 1
x_pred <- seq(a, b, length.out = 100)

# Choose the number of iterations
Niter <- 7500
burnin <- 2500
thin <- 1
samp_size <- (Niter - burnin)/thin

# Run the algorithm
fit_ranef_spl <- bin_logit_randspline(learnCurve$succ, B_basis(learnCurve$time, knots),
                                      B_basis(learnCurve$time, knots), learnCurve$ind,
                                      learnCurve$ni, learnCurve$time,
                                      B_basis(x_pred, knots),
                                      B_basis(x_pred, knots), hypers, Niter,
                                      burnin, thin)


# Compute the quantiles for the estimated mean function
pi_quant_re <- apply(fit_ranef_spl$pi, 2, quantile, prob = c(0.025, 0.975))
mu_quant_re <- apply(fit_ranef_spl$mu, 2, quantile, prob = c(0.025, 0.975))
sig_u_quant_re <- apply(sqrt(fit_ranef_spl$sigma2_u), 2, quantile, prob = c(0.025, 0.975))

u_quant <- apply(fit_ranef_spl$u, c(1, 2), quantile, prob = c(0.025, 0.975))
u_mean <- apply(fit_ranef_spl$u, c(1, 2), mean)

time <- 1
u_summary <- data.frame(ind = factor(1:n),
                        lo = u_quant[2,time,],
                        md = u_mean[time,],
                        hi = u_quant[1,time,])

u_summary <- u_summary[order(u_summary$md),]
u_summary$ind <- factor(u_summary$ind, levels = u_summary$ind)

# Plot the individual random intercepts
ggplot(u_summary, aes(ymin = lo, ymax = hi, x = ind)) +
  geom_linerange(size = 1) +
  xlab("Individual") +
  ylab("Random effects distributions") +
  geom_hline(yintercept = 0, linetype = "dashed") +
  geom_point(aes(y = md), col = cols[1], size = 2) +
  scale_color_brewer(name="", palette = "Set1") +
  ylim(-3, 3) +
  theme(plot.title = element_text(hjust = 0.5))


# Plot the population curve on the probability scale
ggplot() +
  geom_point(aes(x = jitter(time), y = succ/ni),
             col = rgb(0.75, 0.75, 0.75, 0.8), pch = 16, size = 2,
             data = learnCurve) +
  geom_line(aes(x = x_pred, y = colMeans(fit_ranef_spl$pi)), col = cols[1], size = 1) +
  geom_ribbon(aes(x = x_pred, ymin = pi_quant_re[1,], ymax = pi_quant_re[2,]), fill = cols[1], alpha = 0.4) +
  labs(x = 'Number of days', y = 'Probability') +
  scale_color_brewer(name="", palette = "Set1") +
  theme(legend.title=element_blank())


# Random effects variance estimate
ggplot() +
  geom_line(aes(x = x_pred, y = colMeans(sqrt(fit_ranef_spl$sigma2_u))), col = cols[1], size = 1) +
  geom_ribbon(aes(x = x_pred, ymin = sig_u_quant_re[1,], ymax = sig_u_quant_re[2,]), fill = cols[1], alpha = 0.4) +
  labs(x = 'Number of days', y = "Random effects s.d.") +
  scale_color_brewer(name="", palette = "Set1") +
  theme(legend.title=element_blank())


# Plot three individual specific curves
idx1 <- 14
idx2 <- 17
idx3 <- 15


ggplot() +
  geom_point(aes(x = time, y = succ/ni),
             col = rgb(0.75, 0.75, 0.75, 0.8), pch = 16, size = 2,
             data = jit_data) +

  geom_line(aes(x = x_pred, y = colMeans(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx1,])))), col = cols[4], size = 1) +
  geom_point(aes(x = jit_data$time[learnCurve$ind == idx1], y = (learnCurve$succ/learnCurve$ni)[learnCurve$ind == idx1]), col = cols[4], pch = 16, size = 2) +
  geom_line(aes(x = jit_data$time[learnCurve$ind == idx1], y = (learnCurve$succ/learnCurve$ni)[learnCurve$ind == idx1]), size = 0.5, col = cols[4]) +
  geom_ribbon(aes(x = x_pred, ymin = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx1,])), 2, quantile, prob = 0.025), ymax = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx1,])), 2, quantile, prob = 0.975)), fill = cols[4], alpha = 0.4) +

  geom_line(aes(x = x_pred, y = colMeans(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx2,])))), col = cols[5], size = 1) +
  geom_point(aes(x = jit_data$time[learnCurve$ind == idx2], y = (learnCurve$succ/learnCurve$ni)[learnCurve$ind == idx2]), col = cols[5], pch = 16, size = 2) +
  geom_line(aes(x = jit_data$time[learnCurve$ind == idx2], y = (learnCurve$succ/learnCurve$ni)[learnCurve$ind == idx2]), size = 0.5, col = cols[5]) +
  geom_ribbon(aes(x = x_pred, ymin = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx2,])), 2, quantile, prob = 0.025), ymax = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx2,])), 2, quantile, prob = 0.975)), fill = cols[5], alpha = 0.4) +

  geom_line(aes(x = x_pred, y = colMeans(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx3,])))), col = 'aquamarine3', size = 1) +
  geom_point(aes(x = jit_data$time[learnCurve$ind == idx3], y = (learnCurve$succ/learnCurve$ni)[learnCurve$ind == idx3]), col = 'aquamarine3', pch = 16, size = 2) +
  geom_line(aes(x = jit_data$time[learnCurve$ind == idx3], y = (learnCurve$succ/learnCurve$ni)[learnCurve$ind == idx3]), size = 0.5, col = 'aquamarine3') +
  geom_ribbon(aes(x = x_pred, ymin = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx3,])), 2, quantile, prob = 0.025), ymax = apply(logit_mat(fit_ranef_spl$mu + t(fit_ranef_spl$u[,idx3,])), 2, quantile, prob = 0.975)), fill = 'aquamarine3', alpha = 0.4) +

  labs(x = 'Number of days', y = 'Probability') +
  scale_color_brewer(name="", palette = "Set1") +
  theme(legend.title=element_blank())

