#include <RcppArmadillo.h>
#include <progress.hpp>
#include <R_ext/Utils.h>
#include <math.h>
// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::depends(RcppProgress)]]


//' Logit transformation of a matrix
//'
//' @param x input matrix
//' @return the element wise logit transformation of x
//' @export
// [[Rcpp::export]]
arma::mat logit_mat(arma::mat x){
  arma::mat out = arma::exp(x) / (arma::exp(x) + 1.0);

  return (out);
}


//' Logit transformation of a vector
//'
//' @param x input vector
//' @return the element wise logit transformation of x
//' @export
// [[Rcpp::export]]
arma::vec logit_vec(arma::vec x){
  arma::vec out = arma::exp(x) / (arma::exp(x) + 1.0);

  return (out);
}


double dinvgamma(double x, double a, double b, bool logd = false){
  double out = 0.0;

  out = a * std::log(b) - lgamma(a) - (a + 1) * std::log(x) - b / x;

  if (logd == false){
    out = std::exp(out);
  }
  return(out);
}


double dmvnormArma(arma::vec x, arma::vec mean, arma::mat sigma,
                   bool logd = false){
  const int n = x.n_elem;
  double out = 0.0;

  arma::mat rooti = arma::trans(arma::inv(trimatu(arma::chol(sigma))));
  double rootisum = arma::sum(log(rooti.diag()));
  double constants = -(static_cast<double>(n)/2.0) * std::log(2.0 * M_PI);

  arma::vec z = rooti * (x - mean);
  out = constants - 0.5 * arma::sum(z%z) + rootisum;

  if (logd == false){
    out = std::exp(out);
  }
  return(out);
}


double dmvnormArma_prec(arma::vec x, arma::vec mean, arma::mat prec,
                        bool logd = false){
  const int n = x.n_elem;
  double out = 0.0;

  double rootisum = arma::sum(log(prec.diag()));
  double constants = -(static_cast<double>(n)/2.0) * std::log(2.0 * M_PI);

  arma::vec z = prec * (x - mean);
  out = constants - 0.5 * arma::sum(z%z) + rootisum;

  if (logd == false){
    out = std::exp(out);
  }
  return(out);
}



arma::mat rmvnormArma(unsigned int n, arma::vec mean, arma::mat sigma){
  int p = sigma.n_cols;
  arma::mat Y = arma::randn(n, p);

  arma::mat out = repmat(mean, 1, n).t() + Y * arma::chol(sigma);;
  return (out);
}



Rcpp::NumericVector rPG(arma::vec b, arma::vec c){

  // Obtain environment containing function
  Rcpp::Environment base("package:pgdraw");

  // Make function callable from C++
  Rcpp::Function pg_r = base["pgdraw"];

  // Call the function and receive its list output
  Rcpp::NumericVector res = pg_r(b, c);

  return (res);
}





//' Construct the covariance matrix P of the smoothness inducing prior for the spline coefficients
//'
//' @param K number of spline knots
//' @return covariance of the smoothness inducing prior (penalizing second differences in beta)
//' @export
// [[Rcpp::export]]
arma::mat P_smooth(unsigned int K){
  arma::mat D = arma::eye<arma::mat>(K, K);
  arma::mat D_diff(K - 1, K);
  arma::mat D_diff2(K - 2, K);

  for (unsigned i = 0; i < K - 1; i++){
    D_diff.row(i) = D.row(i+1) - D.row(i);
  }
  for (unsigned i = 0; i < K - 2; i++){
    D_diff2.row(i) = D_diff.row(i+1) - D_diff.row(i);
  }

  arma::mat P = D_diff2.t() * D_diff2;

  return (P);
}



//' Spline basis
//'
//' Construct the J basis functions for the splines (evaluated on a fine grid)
//'
//' @param xgrid grid where we want to evaluate the spline functions
//' @param knots vector of knots for the splines
//' @return \eqn{(T \times K+1)} - matrix representing the value of each basis function
//' @export
// [[Rcpp::export]]
arma::mat B_basis(arma::vec xgrid, arma::vec knots){
  unsigned int n = xgrid.n_elem;
  // number of knots
  unsigned int K = knots.n_elem;

  arma::mat B(n, K + 1, arma::fill::zeros);

  arma::uvec act_idx;
  double act_x, resc_x;
  for (unsigned int j = 0; j < K - 1; j++){
    act_idx = find((xgrid >= knots[j]) && (xgrid <= knots[j+1]));

    for (unsigned int k = 0; k < act_idx.n_elem; k++){
      act_x = xgrid(act_idx(k));
      resc_x = (act_x - knots(j)) / (knots(j+1) - knots(j));

      B(act_idx(k),j) = 0.5 * std::pow(1 - resc_x, 2.0);
      B(act_idx(k),j+1) = - std::pow(resc_x, 2.0) + resc_x + 0.5;
      B(act_idx(k),j+2) = 0.5 * std::pow(resc_x, 2.0);
    }
  }

  return(B);
}



//' Functional Logistic Mixed Effects Model
//'
//' Binomial logistic nonparametric model with random effects, fitting a global curve \eqn{mu(t)} on the logit scale via penalized B-splines plus a random effect, i.e.
//'     \eqn{logit(\pi_i(t)) = \mu(t) + \xi(t) \eta_i}
//'
//' @param y vector of observations (number of observed successes for each individual n = 1, ..., N, for each time point t = 1, ..., T)
//' @param X matrix of fixed covariates (in the spline case, it is just the vector of time points corresponding to y expanded on the Spline basis)
//' @param Z matrix of random effects (in the spline case, it is just the vector of time points corresponding to y expanded on the Spline basis)
//' @param ind vector of identifiers for the individuals, corresponding to the observations y (ind_i == j means that y_i is associated to the individual j)
//' @param ni vector of number of trials
//' @param time vector of time indicators for each observation y
//' @param x_grid matrix of the grid where to evaluate the function \eqn{\mu(t)} expanded on the Spline basis)
//' @param z_grid matrix of the grid where to evaluate the function \eqn{\xi(t)} expanded on the Spline basis)
//' @param hypers hyperparameters (P, a_beta, b_beta)
//' @param Niter number of iterations
//' @param burnin burnin length
//' @param thin thinning factor
//' @return object with MCMC output
//' * `beta`: posterior distribution of the spline coefficients
//' * `sigma2_b`: variance of the betas (smoothness penalty)
//' * `pi`: curve on the probability scale (evaluated on x_grid)
//' * `mu`: curve on the logit scale (evaluated on x_grid)
//' * `sigma2_u`: variance of the random effects
//' * `u`: random intercept curves for each individual
//' @export
// [[Rcpp::export]]
SEXP bin_logit_randspline(arma::vec y, arma::mat X, arma::mat Z, arma::uvec ind,
                          arma::vec ni, arma::vec time, arma::mat x_grid,
                          arma::mat z_grid, Rcpp::List hypers,
                          unsigned int Niter, unsigned int burnin,
                          unsigned int thin){
  arma::uvec unique_ind = arma::unique(ind);

  unsigned int n_ind = unique_ind.n_elem; // number of individuals in the study
  unsigned int n = X.n_rows; // n_x is here the dimension of the data (number of individuals * time steps)
  unsigned int p_x = X.n_cols; // p_x is the dimension of the spline basis
  unsigned int p_z = Z.n_cols; // p_z is the dimension of the spline basis
  unsigned int n_grid = x_grid.n_rows;

  unsigned int samp_size = (Niter - burnin)/thin;
  unsigned int it = 0;


  // Define data structures
  arma::mat beta_chain(samp_size, p_x);
  arma::mat gamma_chain(samp_size, p_z);
  arma::mat pi_chain(samp_size, n_grid);
  arma::mat mu_chain(samp_size, n_grid);
  arma::mat eta_chain(samp_size, n_ind);
  arma::cube u_chain(n_grid, n_ind, samp_size);
  arma::vec sig2_b_chain(samp_size);
  arma::vec sig2_gamma_chain(samp_size);
  arma::mat sig2_u_chain(samp_size, n_grid);
  arma::vec log_lik(samp_size);
  arma::vec log_post(samp_size);


  // Initialize data structures
  arma::vec beta_old(p_x, arma::fill::ones);
  arma::vec gamma_old(p_z, arma::fill::ones);
  arma::vec omega_old(n, arma::fill::ones);
  omega_old *= 25;
  arma::vec eta_old(n_ind, arma::fill::zeros);
  double sig2_b_old = 1.0;
  double sig2_eta_old = 1.0;
  double sig2_gamma_old = 1.0;

  // Hyperparameters
  arma::mat P_x = hypers["P_x"];
  arma::mat P_z = hypers["P_z"];
  double a_beta = hypers["a_beta"];
  double b_beta = hypers["b_beta"];
  double a_eta = hypers["a_eta"];
  double a_gamma = hypers["a_gamma"];
  double b_gamma = hypers["b_gamma"];

  // Updated hyperparameters
  double a_beta_star = a_beta + 0.5 * p_x;
  double b_beta_star;
  double a_gamma_star = a_gamma + 0.5 * p_z;
  double b_gamma_star;
  double a_eta_star = 0.5 * (a_eta + n_ind);
  double b_eta_star;

  // Scaled response
  arma::vec kappa = y - 0.5 * ni;

  // Auxiliary variables
  const double c_sq = std::pow((16.0 * std::pow(3.0, 0.5))/(15.0 * M_PI), 2.0);
  arma::vec m_omega;
  arma::mat V_omega;
  arma::vec X_beta(n, arma::fill::zeros);
  arma::vec X_beta_grid(n_grid);
  arma::vec Z_gamma(n, arma::fill::zeros);
  arma::vec Z_gamma_grid(n_grid);
  arma::vec res(n);
  arma::vec rep_eta(n, arma::fill::zeros);
  double post_var, post_mean;
  arma::uvec idx_i;
  arma::vec corr;
  arma::vec zeros_px(p_x, arma::fill::zeros);
  arma::vec zeros_pz(p_z, arma::fill::zeros);
  arma::vec zeros_n(n_ind, arma::fill::zeros);


  Progress bar(Niter, true);
  // Run the Gibbs Sampler
  for (unsigned int iter = 1; iter <= Niter; iter++){
    // (1) Sample new coefficients beta
    res = kappa - Z_gamma % (omega_old % rep_eta);
    V_omega = arma::inv_sympd(X.t() * arma::diagmat(omega_old) * X +
      P_x/sig2_b_old);
    m_omega = V_omega * (X.t() * res);
    beta_old = rmvnormArma(1, m_omega, V_omega).t();


    // (2) Sample the polya gamma auxiliary variables
    X_beta = X * beta_old;
    omega_old = rPG(ni, X_beta + Z_gamma % rep_eta);

    // (3) Sample the new random effects eta_i
    for (unsigned int i = 0; i < n_ind; i++){
      idx_i = find(ind == (i + 1));
      post_var = 1.0 / (1.0/sig2_eta_old + arma::accu(arma::pow(Z_gamma.elem(idx_i), 2.0) % omega_old.elem(idx_i)));
      post_mean = post_var * arma::accu(kappa.elem(idx_i) % Z_gamma.elem(idx_i) - omega_old.elem(idx_i) % X_beta.elem(idx_i) % Z_gamma.elem(idx_i));
      eta_old(i) = R::rnorm(post_mean, std::pow(post_var, 0.5));
      rep_eta.elem(idx_i).zeros();
      rep_eta.elem(idx_i) += eta_old(i);
    }


    // (4) Sample the new xi(t) via the gamma parameter
    res = kappa % rep_eta - X_beta % (omega_old % rep_eta);
    V_omega = arma::inv_sympd(Z.t() * arma::diagmat(omega_old % arma::pow(rep_eta, 2.0)) * Z +
      P_z/sig2_gamma_old);
    m_omega = V_omega * (Z.t() * res);
    gamma_old = rmvnormArma(1, m_omega, V_omega).t();
    Z_gamma = Z * gamma_old;


    // (5) Sample variance of the smoothing spline coefficient sigma_beta
    b_beta_star = arma::as_scalar((b_beta + 0.5 * beta_old.t() * P_x * beta_old)[0]);
    sig2_b_old = 1.0 / arma::as_scalar(Rcpp::rgamma(1, a_beta_star, 1.0/b_beta_star)[0]);


    // (6) Sample variance of the smoothing spline coefficient sigma_gamma
    b_gamma_star = arma::as_scalar((b_gamma + 0.5 * gamma_old.t() * P_z * gamma_old)[0]);
    sig2_gamma_old = 1.0 / arma::as_scalar(Rcpp::rgamma(1, a_gamma_star, 1.0/b_gamma_star)[0]);


    // (7) Sample variance of the random effects sigma_eta
    b_eta_star = 0.5 * (a_eta + arma::accu(arma::pow(eta_old, 2.0)));
    sig2_eta_old = 1.0 / arma::as_scalar(Rcpp::rgamma(1, a_eta_star, 1.0/b_eta_star)[0]);

    if ( (iter > burnin) & (iter % thin == 0) ){
      // Save the values in the chain
      Z_gamma_grid = z_grid * gamma_old;
      X_beta_grid = x_grid * beta_old;
      corr = arma::pow(1.0 + c_sq * arma::pow(Z_gamma_grid, 2.0) * sig2_eta_old, 0.5);
      mu_chain.row(it) = (X_beta_grid).t();
      pi_chain.row(it) = (exp(X_beta_grid / corr)/(1.0 + exp(X_beta_grid / corr))).t();
      sig2_b_chain(it) = sig2_b_old;
      sig2_gamma_chain(it) = sig2_gamma_old;
      sig2_u_chain.row(it) = arma::pow((Z_gamma_grid).t(), 2.0) * sig2_eta_old;
      for (unsigned int i = 0; i < n_ind; i++){
        u_chain.slice(it).col(i) = Z_gamma_grid * eta_old(i);
      }
      beta_chain.row(it) = beta_old.t();
      gamma_chain.row(it) = gamma_old.t();
      eta_chain.row(it) = eta_old.t();


      // Compute log-likelihood and log-posterior
      log_lik(it) = arma::accu(y % (X_beta + Z_gamma % rep_eta) -
        ni % (arma::log(1.0 + exp(X_beta + Z_gamma % rep_eta))));

      log_post(it) = log_lik(it) +
        dmvnormArma_prec(beta_old, zeros_px, sig2_b_old * P_x, true) +
        dinvgamma(sig2_b_old, a_beta, b_beta, true) +
        dmvnormArma_prec(gamma_old, zeros_pz, sig2_gamma_old * P_z, true) +
        dinvgamma(sig2_gamma_old, a_gamma, b_gamma, true) +
        dmvnormArma(eta_old, zeros_n,
                    sig2_eta_old * arma::eye<arma::mat>(n_ind, n_ind), true) +
                      dinvgamma(sig2_eta_old, a_eta/2.0, a_eta/2.0, true);

      it++;
      Rcpp::checkUserInterrupt();
    }
    bar.increment();
  }

  return Rcpp::List::create(Rcpp::Named("beta")=beta_chain,
                            Rcpp::Named("gamma")=gamma_chain,
                            Rcpp::Named("u")=u_chain,
                            Rcpp::Named("sigma2_b")=sig2_b_chain,
                            Rcpp::Named("sigma2_gamma")=sig2_gamma_chain,
                            Rcpp::Named("sigma2_u")=sig2_u_chain,
                            Rcpp::Named("mu")=mu_chain,
                            Rcpp::Named("pi")=pi_chain,
                            Rcpp::Named("eta")=eta_chain,
                            Rcpp::Named("log_lik")=log_lik,
                            Rcpp::Named("log_post")=log_post);
}

